/***************************************************************************
 *   Copyright (C) 2004 by Thomas Nagy                                     *
 *   tnagy2^8@yahoo.fr                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include <stdlib.h>

#include <qsimplerichtext.h>
#include <qpainter.h>
#include <qcolor.h>
#include <qpixmap.h>

#include <klocale.h>
#include <kdebug.h>

#include "constants.h"
#include "datastruct/constants.h"
#include "settings.h"
#include "datastruct/DDataItem.h"
#include "datastruct/DDataControl.h"
#include "itemconfig/ItemCfg.h"
#include "DCanvasView.h"
#include "DCanvasItem.h"
#include "DCanvasLink.h"
#include "DCanvasPos.h"

static const int offset = 4;
static const int maxwidth = 50;
static const int itemz = 5;
static const int linkz = 4;

DCanvasItem::DCanvasItem(DCanvasView * parent, int id) : 
    QCanvasRectangle(parent->canvas()), DGuiItem( parent, id)
{
    m_link = new DCanvasLink( canvas() );
    m_link->setZ(linkz);
    m_link->show();

    m_pos = new DCanvasPos( canvas(), this );
    //m_pos->hide();
    m_pos->setZ(itemz);

    m_canvasview = static_cast<DCanvasView*> (parent);
    setSize(50, 50);
    setSelected(false);
    setDisplayIdx(false);

    updatePos();
    show();
}

DCanvasItem::~DCanvasItem()
{
    hide();
    m_link->hide();
    m_pos->hide();

    delete m_link;
    delete m_pos;
}

int DCanvasItem::rtti() const
{
    return Rtti_DCanvasItem;
}

void DCanvasItem::setSelected(bool b)
{
    // kdWarning()<<"DCanvasItem::setSelected - "<<Id()<<" "<<b<<endl;
    if (b)
	setZ(itemz+1);
    else
	setZ(itemz);
    QCanvasRectangle::setSelected(b);
}

void DCanvasItem::setOffset(const QPoint & p)
{
    m_pressOffset.setX( (int) (p.x() - x()) );
    m_pressOffset.setY( (int) (p.y() - y()) );
}

void DCanvasItem::moveOffset(const QPoint & p)
{
    int newX = p.x() - m_pressOffset.x();
    int newY = p.y() - m_pressOffset.y();

    newX = newX<0 ? 0 : newX;
    newY = newY<0 ? 0 : newY;

    newX = (newX + width()  > canvas()->width() ) ? ( canvas()->width() - width()   ) : newX;
    newY = (newY + height() > canvas()->height()) ? ( canvas()->height() - height() ) : newY;

    move(newX, newY);
}

void DCanvasItem::move(double xx, double yy)
{
    if (xx<0) xx=0;
    if (yy<0) yy=0;
    if (xx>canvas()->width()-width()) xx=canvas()->width()-width();
    if (yy>canvas()->height()-height()) yy=canvas()->height()-height();
    
    QCanvasItem::move(xx, yy);
    m_pos->updatePos();
    m_pos->move( x() - m_pos->width() - 2, y() );

    if (!data())
	return;
    else
	data()->setXY( (int) x(), (int) y());

    updateLink();
}

void DCanvasItem::updateLink()
{
    if ( data()->Parent() != NOITEM )
    {
	DCanvasItem* parn = m_canvasview->canvasItem( data()->Parent() );

	if (!parn) // might be updating the datastructure
	    return;

	if (Settings::useSplines())
	{
	    int diffx=0, diffy=0;
	    if ( ((int) parn->x()) > 10 + ((int) x()) + width() )
	    {
		// parent is on the left
		diffx = ((int) parn->x()) - ( ((int) x()) + width() );
		m_link->setPoints( ((int) x()) + width(), ((int) y()) + height()/2,
			((int) parn->x()) - (1*diffx)/3, ((int) y()) + height()/2,
			((int) parn->x()) - (2*diffx)/3, ((int) parn->y()) + parn->height()/2,
			((int) parn->x()), ((int) parn->y()) + parn->height()/2
			);
	    }
	    else if ( ((int) parn->x()) + parn->width() + 10 < ((int) x()) )
	    {
		// parent is on the right
		diffx = ((int) x()) - ( ((int) parn->x()) + parn->width() );
		m_link->setPoints( ((int) x()), ((int) y()) + height()/2, 
			((int) x()) - (2*diffx)/3, ((int) y()) + height()/2,
			((int) x()) - (1*diffx)/3, ((int) parn->y()) + parn->height()/2,
			((int) parn->x()) + parn->width(), ((int) parn->y()) + parn->height()/2 
			);
	    }
	    else if ( ((int) parn->y()) + parn->height() + 10 < ((int) y()) )
	    {
		// parent is below
		diffy = ((int) y()) - ((int) parn->y()) - parn->height();
		m_link->setPoints( ((int) x()) + width()/2, ((int) y()), 
			((int) x()) + width()/2,             ((int) y()) - (2*diffy)/3,
			((int) parn->x()) + parn->width()/2, ((int) y()) - (1*diffy)/3,
			((int) parn->x()) + parn->width()/2, ((int) parn->y()) + parn->height()
			);
	    }
	    else if ( ((int) parn->y()) > ((int) y()) + height() + 10 )
	    {
		// parent is above
		diffy = ((int) parn->y()) - ((int) y()) - height();
		m_link->setPoints( ((int) x()) + width()/2, ((int) y()) + height(), 
			((int) x()) + width()/2,             ((int) parn->y()) - (1*diffy)/3,
			((int) parn->x()) + parn->width()/2, ((int) parn->y()) - (2*diffy)/3,
			((int) parn->x()) + parn->width()/2, ((int) parn->y())
			);
	    }
	    else
	    {
		// when we dont know how to link, use straight lines
		m_link->setPoints( ((int) x()) + width()/2, ((int) y()) + height()/2, 
			((int) x()) + width()/2, ((int) y()) + height()/2,
			((int) parn->x()) + parn->width()/2, ((int) parn->y()) + parn->height()/2,
			((int) parn->x()) + parn->width()/2, ((int) parn->y()) + parn->height()/2 
			);
	    }
	}
	else
	{
	    // use straight lines
	    m_link->setPoints( ((int) x()) + width()/2, ((int) y()) + height()/2,
		    ((int) x()) + width()/2, ((int) y()) + height()/2,
		    ((int) parn->x()) + parn->width()/2, ((int) parn->y()) + parn->height()/2,
		    ((int) parn->x()) + parn->width()/2, ((int) parn->y()) + parn->height()/2
		    );
	}
    }
    else
    {
	m_link->setPoints(0,0,  0,0,  0,0,  0,0);
    }
}

void DCanvasItem::updatePos()
{
    //kdWarning()<<"updatepos called"<<endl;

    DDataItem *item = data();

    // compute the dimensions first
    QSimpleRichText rt(DDataItem::protectXML(item->dispmsg()), item->m_defaultFont);
    rt.adjustSize();

    m_link->updateColor();

    int pixwidth = 0, pixheight = 0; // dimensions of the canvas item
    
    pixwidth = MAX( rt.width(), item->m_pixtb.width() ) + 2*offset;
    if (!Settings::picturesOnCanvas())
	pixwidth = MAX( rt.width(), item->m_pixtb.width() ) + 2*offset;

    pixheight = item->m_pixtb.height() + rt.height() + 2*offset;
    if (! Settings::picturesOnCanvas())
	pixheight = rt.height() + 2*offset;
    
    setSize( pixwidth, pixheight);
    m_displayedpix.resize( pixwidth, pixheight);

    // set the painter
    QPainter p(&m_displayedpix);
    p.setFont( item->m_defaultFont );

    QColorGroup cg;

    cg.setColor( QColorGroup::Background, item->fillColor() );
    if (Settings::reverseVideo())
    {
	cg.setColor( QColorGroup::Background, QColor(Qt::black) );
	cg.setColor( QColorGroup::Text, item->fillColor() );
	p.setPen( QColor(Qt::white) );
	p.setBrush( QColor(Qt::black) );
    }
    else
    {
	p.setBrush( item->fillColor() );
	p.setPen( item->outlineColor() );
    }

    // paint
    p.drawRoundRect( 0 , 0 , pixwidth, pixheight, 0, 0 );
    
    if ( Settings::picturesOnCanvas() )
	p.drawPixmap( QPoint((pixwidth - item->m_pixtb.width())/2, offset) , item->m_pixtb);

   
    int offsetheight = offset;
    if ( Settings::picturesOnCanvas() )
	offsetheight = item->m_pixtb.height();

    rt.draw(
	    &p,
	    offset+1,
	    offsetheight,
	    QRect( 0, 0, width(), height() ),
	    cg);

    if (data()->Parent() == NOITEM 
	    && data()->colorScheme() != Settings::EnumColorMode::root_ 
	    && data()->colorScheme() != Settings::EnumColorMode::default_)
    {
	p.setBrush( data()->realFillColor() );
	p.drawEllipse(
		width() - 12, 5,
		7, 7);
    }

    // set the position
    if (item)
	move(item->x(), item->y());
    else
	move(x(), y());
    //m_pos->updatePos();
    //m_pos->move( x() - m_pos->width() - 2, y() );
    update();
}

void DCanvasItem::loadPos()
{
    move(data()->x(), data()->y());
}

bool DCanvasItem::onWidget(const QPoint & p)
{
    return ( x() <= p.x() && x() + width() >= p.x() &&
	    y() <= p.y() && y() + height() >= p.y() );
}

void DCanvasItem::drawShape(QPainter &p)
{
    if (!data())
	return;

    drawContent(p);
    if (isSelected())
	drawSelected(p);
}

void DCanvasItem::drawContent(QPainter &p)
{
    if (!data())
    {
	kdWarning()<<"BUG in DCanvasItem::drawContent"<<endl;
	return;
    }

    if (! m_displayedpix.isNull())
    {
	// DEBUG
	//kdWarning()<<"normal behaviour"<<endl;
	p.drawPixmap( (int) x(), (int) y(), m_displayedpix, 0, 0, width(), height());
    }
    else
    {
	p.setBrush( data()->fillColor() );
	p.setPen( data()->outlineColor() );
	p.drawRoundRect( (int) x() , (int) y() , width(), height(), 0, 0 );
    }
}

void DCanvasItem::drawSelected(QPainter& p)
{
    // draw 4 small squares, meaning that the object is selected
    int w = width();
    int h = height();
    int s = 4;

    QBrush brush(blue);
    p.fillRect((int) x()        , (int) y(),         (int) s, s, brush);
    p.fillRect((int) x()        , (int) y() + h - s, (int) s, s, brush);
    p.fillRect((int) x() + w - s, (int) y(),         (int) s, s, brush);
    p.fillRect((int) x() + w - s, (int) y() + h - s, (int) s, s, brush);
}

bool DCanvasItem::displayChildIdx() const
{
    return m_displaychildidx;
}

void DCanvasItem::setDisplayIdx(bool b)
{
    m_displaychildidx = b;
    m_pos->setVisible(b);
}

int DCanvasItem::numChild()
{
    if (data()->Parent() == NOITEM)
	return NOITEM;

    return m_canvasview->dataItem(data()->Parent())->childIdx(Id());
}

void DCanvasItem::setSize( int w, int h )
{
    if (width() != w || height() != h)
    {
	// kdWarning()<<"DCanvasItem::setSize true"<<endl;

	QCanvasRectangle::setSize( w, h );
	//updatePos();
	m_canvasview->dataTree()->notifyChildren(Id());
    }
}

void DCanvasItem::expandSubtree(double angleStart, double angleEnd)
{
    double totalsize = (double) data()->countFamily();

    // find the item angles
    // give an amount of space proportional to the amount of children
    double substart=angleStart, subend=0;
    for (int i=0; i<data()->countChildren(); i++)
    {
	int subsize = data( data()->childNum(i) )->countFamily();

	subend = substart + (subsize * (angleEnd-angleStart)) / totalsize;

	DCanvasItem *citem = m_canvasview->canvasItem( data()->childNum(i) );
    
	citem->m_angle = substart + (subend - substart)/2.;

	citem->move( x() + width()/2. - citem->width()/2. + 130*cos(citem->m_angle), 
		y() + height()/2. - citem->height()/2. + 130*sin(citem->m_angle) );
	
	citem->expandSubtree(substart, subend);
	
	substart = subend;
    }
}

int DCanvasItem::sizeH()
{
    if ( data()->countChildren() > 0 )
    {
	int size = 0;
	for (int i=0; i<data()->countChildren(); i++)
	{
	    DCanvasItem *citem = m_canvasview->canvasItem( data()->childNum( i ) );
	    size += citem->sizeH();
	}
	return size;
    }

    return height() + NESTING;
}

void DCanvasItem::setPosition(int xx, int yy, bool orientation)
{
    // yy is the start of the block
    move(xx, yy+(sizeH()-NESTING)/2-height()/2);

    if ( data()->countChildren() <= 0 )
	return;

    // now the children
    int tmpsize = sizeH();
    int starty = yy;

    for (int i=0; i<data()->countChildren(); i++)
    {
	DCanvasItem *citem = m_canvasview->canvasItem( data()->childNum( i ) );

	if (!citem)
	{
	    kdWarning()<<"bug DCanvasItem::setPosition "<<endl;
	    return;
	}

	int xpos = 0;
	if (orientation)
	    xpos = x()+width()+40;
	else
	    xpos = x()-citem->width()-40;

	citem->setPosition(xpos, starty, orientation);
	starty += citem->sizeH();
    }
}

