/***************************************************************************
 *   Copyright (C) 2004 by Thomas Nagy                                     *
 *   tnagy2^8@yahoo.fr                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qdir.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qmap.h>
#include <qpixmap.h>

#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>

#include <kio/job.h>
#include <kio/netaccess.h>
#include <kzip.h>
#include <kurl.h>
#include <klocale.h>
#include <kdebug.h>
#include <ktempdir.h>
#include <kgenericfactory.h>

#include "datastruct/constants.h"
#include "datastruct/DItem.h"
#include "datastruct/DDataControl.h"
#include "datastruct/DDataItem.h"

#include "kdissOOOdoc.h"

K_EXPORT_COMPONENT_FACTORY(libkdissOOOdoc, kdissOOOdocFactory) 
    QObject *kdissOOOdocFactory::createObject(QObject *, const char *, const char *, const QStringList &) { return new kdissOOOdoc; }

static const int maxpicsize = 700;

kdissOOOdoc::kdissOOOdoc() : DGenerator()
{

}

kdissOOOdoc::~kdissOOOdoc()
{

}

QString kdissOOOdoc::fullName()
{
    return "kdissOOOdoc";
}

QString kdissOOOdoc::group()
{
    return "OpenOffice.org";
}

QString kdissOOOdoc::description()
{
    return i18n("This template generates a simple OpenOffice.org document that can be viewed using any OpenOffice.org file format reader like OpenOffice.org writer or kword");
}

QString kdissOOOdoc::quickstart()
{
    return i18n("Go into the directory created and open the .sxw file generated with oowriter or kword\n");
}

void outputPixParagraph(DDataItem* item, QTextStream & s)
{
    if (item->m_pix.isNull())
	return;

    int width = item->m_pix.width();
    int height = item->m_pix.height();

    int newwidth = width;
    int newheight = width;

    if (width > maxpicsize)
    {
	newwidth = maxpicsize;
	newheight = height*newwidth/width;
    }
    else if (height > maxpicsize)
    {
	newheight = maxpicsize;
	newwidth = width*newheight/height;
    }

    // TODO : this item caption is not perfect
    s<<"<p>\n";
    s<<"<img src=\"pics/"<<item->m_picfilename<<
	"\" alt=\"pic-"<<QString::number(item->Id())<<
	"\" height=\""<<newheight<<"\" width=\""<<newwidth<<"\" />";
    s<<"\n</p>\n";

    s<<"<p>\n";
    if (item->m_piccaption.length() > 0)
	s<<DDataItem::protectXML(item->m_piccaption);
    s<<"\n</p>\n";
}

void outputURL(DDataItem* item, QTextStream & s)
{
    KURL url(item->m_url);

    if (! url.isValid())
	return;
	
    // XXX i think it is not necessary to escape the url - '&' chars are allowed here
    s<<"<p>\n";
    s<<" <a href=\"";
    s<<item->m_url;
    s<<"\" >";
    s<<DDataItem::protectXML(item->m_urlcaption);
    s<<"</a>\n";
    s<<"</p>\n";
}

void outputComment(DDataItem* item, QTextStream & s)
{
    if (item->m_comment.length()>0)
	s<<"<!--"<<item->m_comment<<"-->\n";
}

void printItem(DDataItem* item, QTextStream & s)
{
    s<<"<p>\n";
    s<<DDataItem::protectXML( wipeTags(item->m_text) );
    s<<"\n</p>\n";

    outputPixParagraph(item, s);
    outputURL(item, s);
    outputComment(item, s);
}

void kdissOOOdoc::outputData(DDataItem* item, QTextStream & s, int level)
{
	// url
	if (item->m_url.length()>0)
	{
	    QString urlcaption;
	    if (item->m_urlcaption.length()>0)
		urlcaption = DDataItem::protectXML(item->m_urlcaption);
	    else
		urlcaption = "url";
	    
	    s<<"<text:p text:style-name=\"Standard\">";
	    s<<"<text:a xlink:type=\"simple\" xlink:href=\"";
	    s<<item->m_url;
	    s<<"\">"<<urlcaption<<"</text:a>\n";
	    s<<"</text:p>\n";
	}
	
	// print the comment
	if (item->m_comment.length()>0)
	{
	    s<<"<text:p text:style-name=\"P67\">";
	    s<<DDataItem::protectXML(item->m_comment);
	    s<<"</text:p>\n";
	}
	
	// process the picture
	if (!item->m_pix.isNull())
	{
	    s<<"<text:p text:style-name=\"P"<<level<<"\">\n";
	    s<<"<draw:text-box draw:style-name=\"fr1\" draw:name=\"Cadre1\" text:anchor-type=\"paragraph\" svg:x=\"1.48cm\" svg:y=\"0.226cm\" svg:width=\"5.606cm\" fo:min-height=\"3.605cm\" draw:z-index=\"0\">\n";

	    s<<"<text:p text:style-name=\"Illustration\">\n";

	    // TODO : set the size of the picture ..
	    s<<"<draw:image draw:style-name=\"fr2\" draw:name=\"Image1\" text:anchor-type=\"paragraph\" svg:x=\"0.004cm\" svg:y=\"0.002cm\" svg:width=\"5.606cm\" style:rel-width=\"100%\" svg:height=\"3.605cm\" style:rel-height=\"scale\" draw:z-index=\"1\" xlink:href=\""<<"#Pictures/"<<item->m_picfilename<<"\" xlink:type=\"simple\" xlink:show=\"embed\" xlink:actuate=\"onLoad\"/>Illustration <text:sequence text:ref-name=\"refIllustration0\" text:name=\"Illustration\" text:formula=\"Illustration+1\" style:num-format=\"1\">1</text:sequence>: "<<DDataItem::protectXML(item->m_piccaption)<<"</text:p>";

	    s<<"</draw:text-box>\n";
	    s<<"</text:p>\n";
	}	
}

void kdissOOOdoc::writeItem(DDataItem* item, QTextStream & s, int level)
{
    DGenerator::writeItem(item, s, level);

    // printitem ?
    
    if (level == 0)
    {
	s<<"<text:p text:style-name=\"Heading\">";
	s<<DDataItem::protectXML(item->m_summary);
	s<<"</text:p>\n";

	s<<"<text:p text:style-name=\"Standard\">";
	s<<i18n("Abstract :")<<" "<<DDataItem::protectXML( wipeTags(item->m_text) );
	s<<"</text:p>\n";

	s<<"<text:p text:style-name=\"Standard\"/>\n";

	outputData(item, s, level);
	
	s<<"<text:ordered-list text:style-name=\"L1\">";
	if (item->countChildren() > 0)
	{
	    for (int i = 0; i<item->countChildren(); i++)
		writeItem((DDataItem*) m_data->Item(item->childNum(i)), s, level+1);
	}
	s<<"</text:ordered-list>\n";
    }
    else if (level <= 10)
    {
	s<<"<text:list-item>\n";
	
	s<<"<text:h text:style-name=\"P"<<level<<"\" text:level=\""<<level<<"\">";
	s<<DDataItem::protectXML(item->m_summary);
	s<<"</text:h>\n";
	
	// the text
	s<<"<text:p text:style-name=\"P66\">";
	s<<DDataItem::protectXML( wipeTags(item->m_text) );
	s<<"</text:p>\n";

	outputData(item, s, level);

	if (item->countChildren() > 0)
	{
	    s<<"<text:ordered-list>\n";
	    for (int i = 0; i<item->countChildren(); i++)
		writeItem((DDataItem*) m_data->Item(item->childNum(i)), s, level+1);
	    s<<"</text:ordered-list>\n";
	}

	s<<"</text:list-item>\n";
    }
    else
    {
	// insane, a documment shouldn't grow *that* big
	
	s<<"<text:p text:style-name=\"Standard\">";
	s<<DDataItem::protectXML(item->m_summary);
	s<<"</text:p>\n";

	s<<"<text:p text:style-name=\"Standard\">";
	s<<DDataItem::protectXML( wipeTags(item->m_text) );
	s<<"</text:p>\n";

	outputData(item, s, level);
    }
}

void kdissOOOdoc::writeItemPic(DDataControl * data, DDataItem *item, QTextStream &m)
{
    if (!item)
	return;

    if (! item->m_pix.isNull())
    {
	if (item->m_picfilename.endsWith(".png", false))
	{
	    m<<"<manifest:file-entry manifest:media-type=\"image/png\" manifest:full-path=\"";
	    m<<"Pictures/"+item->m_picfilename;
	    m<<"\"/>\n";
	}
	else if (item->m_picfilename.endsWith(".jpeg", false) || item->m_picfilename.endsWith(".jpg", false))
	{
	    m<<"<manifest:file-entry manifest:media-type=\"image/jpeg\" manifest:full-path=\"";
	    m<<"Pictures/"+item->m_picfilename;
	    m<<"\"/>\n";
	}
	// FIXME : can't remember why ?
	//item->m_pix.save(path+"/Pictures/"+"pic-"+QString::number(data->idNum(i))+".png", "PNG");
    }

    for (int i=0; i<item->countChildren(); i++)
    {
	writeItemPic( data, (DDataItem*) data->Item(item->childNum(i)), m );
    }
}

void kdissOOOdoc::writePics(DDataControl * data, QString path)
{
    // ----------------------------------------------------
    // write the pics and process the META-INF directory
    
    DDataItem *root = (DDataItem*) data->Item( data->rootID() );
    
    // copy the pictures
    KIO::NetAccess::dircopy( KURL(data->getTmpDir()->absPath()), KURL(path+"/Pictures"), NULL );
    
    // proces the manifest file
    QFile manifestfile(path+"/META-INF/manifest.xml");
    if (!manifestfile.open(IO_WriteOnly | IO_Append) )
    {
        kdError()<<"could not open the manifest xml file for append :-/"<<endl;
        return;
    }
    
    QTextStream m(&manifestfile);
    m.setEncoding(QTextStream::UnicodeUTF8);
 
    // write the pictures contained in the subtree considered
    writeItemPic( data, root, m );
    
    // close the manifest file
    m<<"</manifest:manifest>";
    manifestfile.close();
}

bool kdissOOOdoc::writeMaindoc(DDataItem *root, const QString & path)
{
    // ----------------------------------------------------
    // process the content
    
    QFile textfile(path+"/content.xml");
    if (!textfile.open(IO_WriteOnly | IO_Append) )
    {
	kdWarning()<<"could not open the main xml file for append :-/"<<endl;
	return false;
    }

    QTextStream u(&textfile);
    u.setEncoding(QTextStream::UnicodeUTF8);

    // test
    //u<<"<text:p text:style-name=\"Standard\">";
    //u<<DDataItem::protectXML(root->m_summary);
    //u<<"</text:p>\n";

    /*outputPixParagraph(root, u);
    outputURL(root, u);
    outputComment(root, u);*/

    // now the central part, with text, pictures, etc ..
    writeItem(root, u, 0);


    // write the end of document
    u<<"</office:body>\n";
    u<<"</office:document-content>";
    
    // close the content document
    textfile.close();

    // ----------------------------------------------------
    // TODO process the meta-data section
    
    // store the author name, title and time creation
    /*struct passwd *pw = ::getpwuid(getuid());
    if (!pw)
	return;
    */

    // title : DDataItem::protectXML(root->m_summary)
    // author : QString::fromLocal8Bit( pw->pw_gecos )
    
    // ----------------------------------------------------
    // create the archive

    KZip zip(path+"/newdoc.sxw");
    zip.setCompression( KZip::DeflateCompression );
    zip.open( IO_WriteOnly );

    zip.addLocalFile(path+"/content.xml", "content.xml");
    zip.addLocalFile(path+"/meta.xml", "meta.xml");
    zip.addLocalFile(path+"/mimetype", "mimetype");
    zip.addLocalFile(path+"/settings.xml", "settings.xml");
    zip.addLocalFile(path+"/styles.xml", "styles.xml");
    zip.addLocalDirectory(path+"/Pictures", "Pictures");
    zip.addLocalDirectory(path+"/META-INF", "META-INF");
    
    zip.close();

    // ----------------------------------------------------
    // remove the (now useless) temporary files

    // do not show the progressdialog
    KIO::del(KURL(path+"/content.xml"), false, false );
    KIO::del(KURL(path+"/meta.xml"), false, false );
    KIO::del(KURL(path+"/mimetype"), false, false );
    KIO::del(KURL(path+"/settings.xml"), false, false );
    KIO::del(KURL(path+"/styles.xml"), false, false );
    KIO::del(KURL(path+"/META-INF"), false, false );
    KIO::del(KURL(path+"/Pictures"), false, false );
    
    return true;
}

#include "kdissOOOdoc.moc"
